//========================================================================
// New gets() clone - if you need one.
// Tested:
// LCC   (Jul 10 2002) -A getline.c
// BCC32 (BCC 5.5.1) -WU -w getline.c
//========================================================================
#include <windows.h>
#include <tchar.h>
#include <stdarg.h>
#include <stdio.h>
#include <string.h>
#include <assert.h>

//========================================================================
// Global variables
//========================================================================
BOOL    bInsert = TRUE;
SHORT   maxx;
SHORT   maxy;

//========================================================================
// Function prototypes
//========================================================================
int _tmain(void);
BOOL PutString(LPTSTR szText);
BOOL GetCursorXY(PSHORT x, PSHORT y);
BOOL GetScreenSize(PSHORT maxx, PSHORT maxy);
BOOL SetCursorType(BOOL bInsert, BOOL bVisible);
BOOL GetKey(PINPUT_RECORD lpBuffer);
SHORT GetCursorX(VOID);
SHORT GetCursorY(VOID);
BOOL SetCursorXY (SHORT x, SHORT y);
BOOL PutChar(TCHAR c);
BOOL ClearLine (LPTSTR str, INT maxlen, SHORT orgx, SHORT orgy);
BOOL JumpLeft(PINT current, LPTSTR str, PSHORT orgx, PSHORT orgy);
BOOL JumpRight(PINT current, LPTSTR str, INT charcount);
BOOL BackSpace(PINT current, PINT charcount, INT maxx, PSHORT curx,
     PSHORT cury, LPTSTR str);
BOOL GetLine(LPTSTR str, INT maxlen);

//========================================================================
// Function..: _tmain
// Purpose...: Where all then fun begin.
// Return....: None
//========================================================================
int _tmain(void)
{
    TCHAR buffer[MAX_PATH];

    if(!GetLine(buffer, MAX_PATH))
        PutString(_T("GetLine() return FALSE!"));

    PutChar(_T('\n'));
    PutString(buffer);
    return 0;
}

#ifdef __LCC__
#define _tcsnset _strnset

//========================================================================
// Function..: strnset (Borland specifid string function).
// Purpose...: Sets a specified number of characters in a string to a
//             given character. strnset copies the character ch into the
//             first n bytes of the string s. If n > strlen(s), then
//             strlen(s) replaces n. It stops when n characters have been
//             set, or when a null character is found.
// Return....: s
//========================================================================
LPTSTR _strnset(LPTSTR s, TCHAR ch, DWORD n)
{
    assert(s != NULL);

    while(*s && n)
    {
        *s = ch;
        n--;
        s++;
   }
   return s;
}
#endif

//========================================================================
// Function..: PutString
// Purpose...: Print string to console.
// Return....: True if success.
//========================================================================
BOOL PutString(LPTSTR szText)
{
    DWORD   dwWritten;

    assert(szText != NULL);

    return WriteConsole(GetStdHandle (STD_OUTPUT_HANDLE),
    szText, lstrlen(szText),&dwWritten, NULL);
}

//========================================================================
// Function..: GetCursorXY
// Purpose...: Get cursor position.
// Return....: True if success.
//========================================================================
BOOL GetCursorXY(PSHORT x, PSHORT y)
{
    CONSOLE_SCREEN_BUFFER_INFO csbi;

    assert(x != NULL);
    assert(y != NULL);

    if(GetConsoleScreenBufferInfo (
        GetStdHandle(STD_OUTPUT_HANDLE), &csbi))
    {
        *x = csbi.dwCursorPosition.X;
        *y = csbi.dwCursorPosition.Y;
        return TRUE;
    }
    else
    {
        *x = 0;
        *y = 0;
        return FALSE;
    }
}

//========================================================================
// Function..: GetScreenSize
// Purpose...: Get console size.
// Return....: True if success.
//========================================================================
BOOL GetScreenSize(PSHORT maxx, PSHORT maxy)
{
    CONSOLE_SCREEN_BUFFER_INFO csbi;

    assert(maxx != NULL);
    assert(maxy != NULL);

    if(GetConsoleScreenBufferInfo (
        GetStdHandle (STD_OUTPUT_HANDLE), &csbi))
    {
        if (maxx) *maxx = csbi.dwSize.X;
        if (maxy) *maxy = csbi.dwSize.Y;
        return TRUE;
    }
    else
    {
        if (maxx) *maxx = 0;
        if (maxy) *maxy = 0;
        return TRUE;
    }
}

//========================================================================
// Function..: SetCursorType
// Purpose...: Set line or black cursor, also control visible state.
// Return....: True if success.
//========================================================================
BOOL SetCursorType(BOOL bInsert, BOOL bVisible)
{
    CONSOLE_CURSOR_INFO cci;

    cci.dwSize = bInsert ? 10 : 99;
    cci.bVisible = bVisible;

    return SetConsoleCursorInfo (GetStdHandle (STD_OUTPUT_HANDLE), &cci);
}

//========================================================================
// Function..: GetKey
// Purpose...: Get key from console.
// Return....: True if success.
//========================================================================
BOOL GetKey(PINPUT_RECORD lpBuffer)
{
    HANDLE hInput = GetStdHandle (STD_INPUT_HANDLE);
    DWORD  dwRead;

    assert(lpBuffer != NULL);

    do{
        if(!ReadConsoleInput (hInput, lpBuffer, 1, &dwRead))
            return FALSE;

    }while(!(lpBuffer->EventType != KEY_EVENT) &&
            (lpBuffer->Event.KeyEvent.bKeyDown != TRUE));

    return TRUE;
}

//========================================================================
// Function..: GetCursorX
// Purpose...: Return cursor X coordinate.
// Return....: -1 if error.
//========================================================================
SHORT GetCursorX(VOID)
{
    CONSOLE_SCREEN_BUFFER_INFO csbi;

    if(!GetConsoleScreenBufferInfo (
        GetStdHandle (STD_OUTPUT_HANDLE), &csbi))
        return -1;

    return csbi.dwCursorPosition.X;
}

//========================================================================
// Function..: GetCursorY
// Purpose...: Return cursor Y coordinate.
// Return....: -1 if error.
//========================================================================
SHORT GetCursorY(VOID)
{
    CONSOLE_SCREEN_BUFFER_INFO csbi;

    if(!GetConsoleScreenBufferInfo (
        GetStdHandle (STD_OUTPUT_HANDLE), &csbi))
        return -1;

    return csbi.dwCursorPosition.Y;
}

//========================================================================
// Function..: GetCursorY
// Purpose...: Set cursor in specifid place.
// Return....: True if success.
//========================================================================
BOOL SetCursorXY (SHORT x, SHORT y)
{
    COORD coPos;

    assert(x > -1);
    assert(y > -1);

    coPos.X = x;
    coPos.Y = y;

    if(!SetConsoleCursorPosition (
        GetStdHandle (STD_OUTPUT_HANDLE), coPos))
        return FALSE;
    else
        return TRUE;
}


//========================================================================
// Function..: PutChar
// Purpose...: Print single charter in console.
// Return....: True if success.
//========================================================================
BOOL PutChar(TCHAR c)
{
    DWORD dwWritten;

    assert(c != 0);

    return WriteConsole(GetStdHandle (STD_OUTPUT_HANDLE),
    &c, 1 ,&dwWritten, NULL);
}

//========================================================================
// Function..: ClearLine
// Purpose...: Clear all charters in line and move cursor to begin.
// Return....: True if success.
//========================================================================
BOOL ClearLine (LPTSTR str, INT maxlen, SHORT orgx, SHORT orgy)
{
    INT count;

    if(!SetCursorXY (orgx, orgy))
        return FALSE;

    for (count = 0; count < (INT)_tcslen (str); count++)
        if(!PutChar (_T(' ')))
            return FALSE;

    _tcsnset (str, _T('\0'), maxlen);
    if(!SetCursorXY (orgx, orgy))
        return FALSE;

    return TRUE;
}

//========================================================================
// Function..: JumpLeft
// Purpose...: Jump to previos word in input line.
// Parameters: current - the position of the cursor in the string (str).
//             str     - string that user is editing
//             orgx    - Cursor orginal x coordinate
//             orgy    - cursor orginal y coordinate
// Return....: True if success.
//========================================================================
BOOL JumpLeft(PINT current, LPTSTR str, PSHORT orgx, PSHORT orgy)
{
    SHORT minus=0;                  // how meny charters need jump
    SHORT nx,ny;                    // new cursor position
    INT c = *current;               // temp counter for new place
    LPTSTR p = str;                 // pointer for searcing jump place
    SHORT t,t2;

    p += *current;                  // jump that position where cursor is

    if(*current <= 0)               // check for begin
        return TRUE;

    while(*p && *p != _T(' '))      // search the previos space charter
    {
        p--;
        c--;
        minus++;
    }

    if(c == -1)                     // it was first word
    {
        if(!SetCursorXY(*orgx,*orgy))   // so we move cursor begin of line
            return FALSE;
        *current = 0;
        return TRUE;
    }

    if(*p == _T(' '))               // there is maby meny spaces
    {
        while(*p && *p == _T(' '))  // jump over them
        {
            p--;
            c--;
            minus++;
        }
        p++;
        c++;
        minus--;
    }

    if(*p == _T('\0'))              // string begin so don't do nothing.
        return TRUE;

    *current = c;                   // save new cursor position
    t = GetCursorX();
    if(t == -1)
        return FALSE;

    if(minus > t)                   // do cursor need move up
    {
        ny = GetCursorY();
        nx = GetCursorX();
        if(ny == -1 || nx == -1)
            return FALSE;

        minus -= nx;                // decrese this line and move up
        ny--;

        while(minus > maxx)         // jump over meny lines if needed
        {
            minus -= maxx;
            ny--;
        }


        nx = maxx;                  // new x coordinate
        nx -= minus;
        if(!SetCursorXY(nx,ny))     // apply
            return FALSE;
    }
    else                            // or more easy way.
    {
        t = GetCursorX();
        if(t == -1) return FALSE;
        t2 = GetCursorY();
        if(t2 == -1) return FALSE;
        t -= minus;
        if(!SetCursorXY(t, t2))
            return FALSE;
    }
    return TRUE;
}

//========================================================================
// Function..: JumpRight
// Purpose...: Jump to next word in input line.
// Parameters: current - the position of the cursor in the string (str).
//             str     - string that user is editing
// Return....: True if success.
//========================================================================
BOOL JumpRight(PINT current, LPTSTR str, INT charcount)
{
    SHORT minus=0;                  // how meny charters need jump
    SHORT nx,ny;                        // new cursor position
    INT c = *current;               // temp counter for new place
    LPTSTR p = str;                 // pointer for searcing jump place
    LPTSTR p2;
    BOOL bEnd=FALSE;
    SHORT t,t2;

    p += *current;                  // jump that position where cursor is

    if(*current == charcount)       // if cursor is end of line do nothing
        return TRUE;

    while(*p && *p != _T(' '))      // search the next space charter
    {
        p++;
        c++;
        minus++;
    }

    if(*p == _T(' '))               // there is maby meny spaces
    {
        while(*p && *p == _T(' '))  // jump over them
        {
            p++;
            c++;
            minus++;
        }
        p--;
        c--;
        minus--;
    }

    p2 = p;
    while(1)
    {
        if(*p2 == _T('\0'))
            bEnd = TRUE;

        if(*p2 == _T(' '))
            break;

        p2++;
    }

    if(bEnd)
    {
        *current = charcount;
    }
    else
        *current = c;                   // save new cursor position

    if(minus > maxx - GetCursorX())     // do cursor need move down
    {
        ny = GetCursorY();
        nx = GetCursorX();

        if(ny == -1 || nx == -1)
            return FALSE;

        minus -= maxx;              // decrese this line and move down
        minus -= nx;
        ny++;

        while(minus > maxx)         // jump over meny lines if needed
        {
            minus -= maxx;
            ny++;
        }

        nx = minus;                 // new x coordinate
        if(!SetCursorXY(nx,ny))     // apply
            return FALSE;
    }
    else                            // or more easy way.
    {
        t = GetCursorX();
        if(t == -1) return FALSE;
        t2 = GetCursorY();
        if(t2 == -1) return FALSE;
        t += minus;
        if(!SetCursorXY(t, t2))
            return FALSE;
    }
    return TRUE;
}

//========================================================================
// Function..: BackSpace
// Purpose...: Delete character to left of cursor */
// Parameters: str - buffer for string.
//             maxlen - size of buffer.
// Globals...: None
// Return....: True if success.
//========================================================================
BOOL BackSpace(PINT current, PINT charcount, INT maxx, PSHORT curx,
PSHORT cury, LPTSTR str)
{
    INT count;
    SHORT x,y;

    if(*current <= 0 && *charcount <= 0)
        return TRUE;

    if (*current == *charcount)
    {
        // if at end of line
        str[*current - 1] = _T('\0');

        x = GetCursorX ();
        if(x == -1) return FALSE;

        if (x != 0)
        {
            if(!PutString(_T("\b \b")))
                return FALSE;
        }
        else
        {
            y = GetCursorY();
            if(y == -1) return FALSE;

            if(!SetCursorXY ((SHORT)(maxx - 1), (SHORT)(y - 1)))
                return FALSE;
            if(!PutChar(_T(' ')))
                return FALSE;

            y = GetCursorY();
            if(y == -1) return FALSE;
            if(!SetCursorXY ((SHORT)(maxx - 1), (SHORT)(y - 1)))
                return FALSE;
        }
    }
    else
    {
        for (count = *current - 1; count < *charcount; count++)
            str[count] = str[count + 1];

        x = GetCursorX();
        if(x == -1) return FALSE;

        if (x != 0)
        {
            x = GetCursorX();
            if(x == -1) return FALSE;
            GetCursorY();
            if(y == -1) return FALSE;

            x--;
            if(!SetCursorXY (x, y))
                return FALSE;
        }
        else
        {
            y = GetCursorY();
            if(y == -1) return FALSE;

            if(!SetCursorXY ((SHORT)(maxx - 1), (SHORT)(y - 1)))
                return FALSE;
        }
        if(!GetCursorXY (curx, cury))
            return FALSE;

        if(!PutString(&str[*current - 1]))
            return FALSE;
        if(!SetCursorXY (*curx, *cury))
            return FALSE;
    }
    --*charcount;
    --*current;

    return TRUE;
}

//========================================================================
// Function..: Getline
// Purpose...: Get string line from console.
// Parameters: str - buffer for string.
//             maxlen - size of buffer.
// Globals...: None
// Return....: True if success.
//========================================================================
BOOL GetLine(LPTSTR str, INT maxlen)
{
    SHORT orgx;             // origin x/y
    SHORT orgy;
    SHORT curx;             // current x/y cursor position
    SHORT cury;
    INT   count;            // used in some for loops
    INT   current = 0;      // position of the cursor in the string (str)
    INT   charcount = 0;    // chars in the string (str)
    TCHAR ch;
    INPUT_RECORD ir;
    SHORT   x,y;

    ZeroMemory(str, maxlen * sizeof (TCHAR));

    if(!GetScreenSize (&maxx, &maxy))
        return FALSE;

    if(!GetCursorXY (&orgx, &orgy))
        return FALSE;

    if(!SetCursorType (bInsert, TRUE))
        return FALSE;

    do{
        if(!GetKey (&ir))
            return FALSE;

        if  (ir.Event.KeyEvent.dwControlKeyState &
            (RIGHT_CTRL_PRESSED|LEFT_CTRL_PRESSED))
        {
            switch (ir.Event.KeyEvent.wVirtualKeyCode)
            {
                case VK_LEFT:
                    if(!JumpLeft(&current, str, &orgx, &orgy))
                        return FALSE;
                break;

                case VK_RIGHT:
                    if(!JumpRight(&current, str, charcount))
                        return FALSE;
                break;
            }
        }

        switch (ir.Event.KeyEvent.wVirtualKeyCode)
        {
            case VK_BACK:
                if(!BackSpace(&current, &charcount, maxx,
                    &curx, &cury, str))
                    return FALSE;
            break;

            case VK_INSERT: // toggle insert/overstrike mode
                bInsert ^= TRUE;
                if(!SetCursorType (bInsert, TRUE))
                    return FALSE;
                break;

            case VK_DELETE: // delete character under cursor.
                if (current != charcount && charcount > 0)
                {
                    for (count = current; count < charcount; count++)
                        str[count] = str[count + 1];
                    charcount--;
                    if(!GetCursorXY (&curx, &cury))
                        return FALSE;
                    if(!PutString(&str[current]))
                        return FALSE;
                    if(!SetCursorXY (curx, cury))
                        return FALSE;
                }
                break;

            case VK_HOME: // goto beginning of string
                if (current != 0)
                {
                    if(!SetCursorXY (orgx, orgy))
                        return FALSE;
                    current = 0;
                }
                break;

            case VK_END: // goto end of string
                if (current != charcount)
                {
                    if(!SetCursorXY (orgx, orgy))
                        return FALSE;
                    if(!PutString(&str[current]))
                        return FALSE;
                    current = charcount;
                }
                break;

            case VK_LEFT: // move cursor left
                if (current > 0)
                {
                    current--;
                    x = GetCursorX();
                    if(x == -1) return FALSE;

                    if (x == 0)
                    {
                        y = GetCursorY();
                        if(y == -1) return FALSE;
                        SetCursorXY ((SHORT)(maxx - 1), (SHORT)(y - 1));
                    }
                    else
                    {
                        x = GetCursorX();
                        if(x == -1) return FALSE;
                        y = GetCursorY();
                        if(y == -1) return FALSE;

                        SetCursorXY ((SHORT)(x - 1), y);
                    }
                }
            break;

            case VK_RIGHT: // move cursor right
                if (current != charcount)
                {
                    current++;
                    x = GetCursorX();
                    if(x == -1) return FALSE;

                    if (x == maxx - 1)
                    {
                        y = GetCursorY();
                        if(y == -1) return FALSE;

                        SetCursorXY (0, (SHORT)(y + 1));
                    }
                    else
                    {
                        y = GetCursorY();
                        if(y == -1) return FALSE;
                        x = GetCursorX();
                        if(x == -1) return FALSE;

                        x++;
                        if(!SetCursorXY(x,y))
                            return FALSE;
                    }
                }
            break;

            case VK_ESCAPE: // clear string
                if(!ClearLine (str, maxlen, orgx, orgy))
                    return FALSE;
                current = charcount = 0;
            break;

            default:
#ifdef _UNICODE
                ch = ir.Event.KeyEvent.uChar.UnicodeChar;
                if ((ch >= 32 && ch <= 255) &&
                    (charcount != (maxlen - 2)))
#else
                ch = ir.Event.KeyEvent.uChar.AsciiChar;

                // Skip 0-shift and 22-win buttons.
                if (ch != 0 && ch != 22 && charcount != (maxlen - 2))
#endif
                {
                    // insert character into string...
                    if (bInsert && current != charcount)
                    {
                        for (count = charcount; count > current; count--)
                            str[count] = str[count - 1];
                        str[current++] = ch;
                        x = GetCursorX();
                        if(x == -1) return FALSE;
                        if (x == maxx - 1)
                        {
                            curx = 0;
                            cury = GetCursorY();
                            cury++;
                            if(cury == 0) return FALSE;
                        }
                        else
                        {
                            if(!GetCursorXY (&curx, &cury))
                                return FALSE;
                            curx++;
                        }

                        if(!PutString(&str[current - 1]))
                            return FALSE;
                        if ((_tcslen (str) > (DWORD)(maxx - orgx)) &&
                            (orgy == maxy + 1))
                                cury--;
                        if(!SetCursorXY (curx, cury))
                            return FALSE;
                        charcount++;
                    }
                    else
                    {
                        if (current == charcount)
                            charcount++;
                        str[current++] = ch;
                        if(!PutChar (ch))
                            return FALSE;
                    }
                    if ((_tcslen (str) > (DWORD)(maxx - orgx)) &&
                        (orgy == maxy + 1))
                        orgy--;
                }
                break;

        }
    }
    while (ir.Event.KeyEvent.wVirtualKeyCode != VK_RETURN);

    return SetCursorType (bInsert, TRUE);
}
